#include <e.h>
#include "e_mod_main.h"
#include "e_effects.h"
#include "e_effects_object.h"

#include <stdio.h>
#include <float.h>

static Ecore_X_Window  input_window = 0;
static Eina_List      *effects_windows = NULL;
static Eina_List      *handlers = NULL;
static Ecore_Animator *anim = NULL;
static double          anim_time = 0.f;
static double          anim_length = 0.33f;
static int             revert = 0;
static Ecore_Evas     *canvas = NULL;
static E_Zone         *ref_zone = NULL;
static Evas           *zone_evas = NULL;

static void _place_windows();
static int _e_effects_cb_key_down(void *data, int type, void *event);
static int _e_effects_cb_mouse_down(void *data, int type, void *event);
static int _e_effects_cb_mouse_wheel(void *data, int type, void *event);
static int _e_effects_animate_cb(void *data);
static int _e_revert_done();

EAPI int
e_effects_init(void)
{
   return 1;
}

EAPI int
e_effects_shutdown(void)
{
   if (effects_windows)
      _e_revert_done();
   return 1;
}

EAPI void
e_effects_apply(E_Zone * zone)
{
   E_Desk *desk;

   E_Border_List *bl;

   E_Border *border;

   if (!ecore_x_composite_query())
      return;
   if (effects_windows)
      return;
   if (canvas)
      return;

   E_OBJECT_CHECK(zone);
   desk = e_desk_current_get(zone);
   bl = e_container_border_list_first(desk->zone->container);

   ref_zone = zone;
   zone_evas = evas_object_evas_get(zone->bg_object);
   e_object_ref(E_OBJECT(zone));
   canvas = e_canvas_new(E_EVAS_ENGINE_SOFTWARE_X11, zone->container->win,
			 zone->x, zone->y, zone->w, zone->h, 1, 1, NULL);

   /* Yank the background from the zone and put it in our app instead
    * of a costly copy :-D. */
   if (zone_evas)
     {
	evas_object_release(zone->bg_object, 0);
	evas_object_inject(zone->bg_object, ecore_evas_get(canvas));
     }

   /* Build the list of affected windows */
   while ((border = e_container_border_list_next(bl)))
     {
	/* Build a list of windows on this desktop and not iconified. */
	if ((border->desk == desk) && (!border->iconic))
	  {
	     Eina_List *ll;

	     E_Effects_Object *fx;

	     int x, y;

	     int w, h;

	     fx = e_effects_object_new(canvas, border,
				       border->x, border->y,
				       border->w, border->h);

	     effects_windows = eina_list_append(effects_windows, fx);
	  }
     }

   _place_windows(zone->w, zone->h);

   /* Animations */
   revert = 0;
   if (!anim)
     {
	anim_time = ecore_loop_time_get();
	anim = ecore_animator_add(_e_effects_animate_cb, effects_windows);
     }

   /* Take over the inputs */
   input_window = ecore_x_window_input_new(zone->container->win, 0, 0, 1, 1);
   ecore_x_window_show(input_window);
   if (!e_grabinput_get(input_window, 0, input_window))
     {
	ecore_x_window_free(input_window);
	input_window = 0;
	e_effects_revert();
	return;
     }

   handlers = eina_list_append
      (handlers, ecore_event_handler_add
       (ECORE_EVENT_KEY_DOWN, _e_effects_cb_key_down, NULL));
   handlers = eina_list_append
      (handlers, ecore_event_handler_add
       (ECORE_EVENT_MOUSE_BUTTON_DOWN, _e_effects_cb_mouse_down, NULL));
   handlers = eina_list_append
      (handlers, ecore_event_handler_add
       (ECORE_EVENT_MOUSE_WHEEL, _e_effects_cb_mouse_wheel, NULL));
}

EAPI void
e_effects_revert()
{
   Eina_List *l;

   E_Effects_Object *data;

   if (!effects_windows)
      return;

   /* Animations */
   EINA_LIST_FOREACH(effects_windows, l, data)
   {
      data->src_x = data->x;
      data->dst_x = data->border->x;
      data->delta_x = (float)(data->dst_x - data->x) / anim_length;
      data->src_y = data->y;
      data->dst_y = data->border->y;
      data->delta_y = (float)(data->dst_y - data->y) / anim_length;
      data->src_w = data->w;
      data->dst_w = data->border->w;
      data->delta_w = (float)(data->dst_w - data->w) / anim_length;
      data->src_h = data->h;
      data->dst_h = data->border->h;
      data->delta_h = (float)(data->dst_h - data->h) / anim_length;
   }
   revert = 1;
   if (!anim)
     {
	anim_time = ecore_loop_time_get();
	anim = ecore_animator_add(_e_effects_animate_cb, effects_windows);
     }
}

static void
_place_windows(int zone_width, int zone_height)
{
   E_Effects_Object *data, *data2;

   Eina_List *l, *l2;

   int desk_x = 0;

   int desk_y = 0;

   int desk_w = zone_width;

   int desk_h = zone_height;

   int done = 0;

   float scale = 1.f;

   float scale_x, scale_y;

   int collision = 1;

   /* Initialize the windows */
   EINA_LIST_FOREACH(effects_windows, l, data)
   {
      e_effects_object_show(data);
      data->dst_x = data->src_x = data->border->x;
      data->dst_y = data->src_y = data->border->y;
      data->dst_w = data->src_w = data->border->w;
      data->dst_h = data->src_h = data->border->h;
   }

   while (!done)
     {
	/* TODO
	 * don't finish until there are no collisions left
	 */
	done = 1;
	EINA_LIST_FOREACH(effects_windows, l, data)
	{
	   EINA_LIST_FOREACH(effects_windows, l2, data2)
	   {
	      collision = 1;

	      /* check for a collision */
	      if (data == data2 ||
		  data->dst_x + data->dst_w <= data2->dst_x ||
		  data->dst_x >= data2->dst_x + data2->dst_w ||
		  data->dst_y + data->dst_h <= data2->dst_y ||
		  data->dst_y >= data2->dst_y + data2->dst_h)
		{
		   collision = 0;
		}

	      /* if there is a collision calculate the centre position
	       * and move the windows to the side closest to the center
	       * of the screen which doesn't have a collision.
	       */
	      if (collision)
		{
		   float shortest_distance_squared = FLT_MAX;

		   float distance_squared;

		   int x, y;

		   int fin_x = data2->dst_x;

		   int fin_y = data2->dst_y;

		   /* check bottom */
		   x = data2->dst_x;
		   y = data->dst_y + data->dst_h;
		   distance_squared = (x - data2->src_x) * (x - data2->src_x)
		      + (y - data2->src_y) * (y - data2->src_y);
		   if (distance_squared < shortest_distance_squared &&
		       (data->dst_x + data->dst_w <= x ||
			data->dst_x >= x + data2->dst_w ||
			data->dst_y + data->dst_h <= y ||
			data->dst_y >= y + data2->dst_h))
		     {
			shortest_distance_squared = distance_squared;
			fin_x = x;
			fin_y = y;
		     }

		   /* check top */
		   x = data2->dst_x;
		   y = data->dst_y - data2->dst_h;
		   distance_squared = (x - data2->src_x) * (x - data2->src_x)
		      + (y - data2->src_y) * (y - data2->src_y);
		   if (distance_squared < shortest_distance_squared &&
		       (data->dst_x + data->dst_w <= x ||
			data->dst_x >= x + data2->dst_w ||
			data->dst_y + data->dst_h <= y ||
			data->dst_y >= y + data2->dst_h))
		     {
			shortest_distance_squared = distance_squared;
			fin_x = x;
			fin_y = y;
		     }

		   /* check right */
		   x = data->dst_x + data->dst_w;
		   y = data2->dst_y;
		   distance_squared = (x - data2->src_x) * (x - data2->src_x)
		      + (y - data2->src_y) * (y - data2->src_y);
		   if (distance_squared < shortest_distance_squared &&
		       (data->dst_x + data->dst_w <= x ||
			data->dst_x >= x + data2->dst_w ||
			data->dst_y + data->dst_h <= y ||
			data->dst_y >= y + data2->dst_h))
		     {
			shortest_distance_squared = distance_squared;
			fin_x = x;
			fin_y = y;
		     }

		   /* check left */
		   x = data->dst_x - data2->dst_w;
		   y = data2->dst_y;
		   distance_squared = (x - data2->src_x) * (x - data2->src_x)
		      + (y - data2->src_y) * (y - data2->src_y);
		   if (distance_squared < shortest_distance_squared &&
		       (data->dst_x + data->dst_w <= x ||
			data->dst_x >= x + data2->dst_w ||
			data->dst_y + data->dst_h <= y ||
			data->dst_y >= y + data2->dst_h))
		     {
			shortest_distance_squared = distance_squared;
			fin_x = x;
			fin_y = y;
		     }

		   if (fin_x < desk_x)
		      desk_x = fin_x;
		   if (fin_y < desk_y)
		      desk_y = fin_y;
		   if (fin_x + data2->dst_w > desk_w)
		      desk_w = fin_x + data2->dst_w;
		   if (fin_y + data2->dst_h > desk_h)
		      desk_h = fin_y + data2->dst_h;

		   data2->dst_x = fin_x;
		   data2->dst_y = fin_y;
		}
	   }
	}
     }

   scale_x = (float)zone_width / (float)(desk_w - desk_x);
   scale_y = (float)zone_height / (float)(desk_h - desk_y);

   if (scale_x < scale_y)
     {
	scale = scale_x;
     }
   else
     {
	scale = scale_y;
     }

   /* set up the animation values */
   EINA_LIST_FOREACH(effects_windows, l, data)
   {
      data->dst_x -= desk_x;
      data->dst_y -= desk_y;
      data->dst_x *= scale;
      data->dst_y *= scale;
      data->dst_w *= scale;
      data->dst_h *= scale;

      data->delta_x = (data->dst_x - data->src_x) / anim_length;
      data->delta_y = (data->dst_y - data->src_y) / anim_length;
      data->delta_w = (data->dst_w - data->src_w) / anim_length;
      data->delta_h = (data->dst_h - data->src_h) / anim_length;

      /* hack so the objects get dirtied and drawn */
      data->src_x++;
   }
}

static int
_e_effects_cb_key_down(void *data, int type, void *event)
{
   Ecore_Event_Key *ev;

   ev = event;
   if (!strcmp(ev->key, "space"))
      e_effects_revert();
   return 1;
}

static int
_e_effects_cb_mouse_down(void *unused, int type, void *event)
{
   Evas *evas;

   Eina_List *l;

   Evas_Object *obj;

   E_Effects_Object *fx;

   E_Effects_Object *fx_under_mouse = NULL;

   Ecore_Event_Mouse_Button *ev;

   evas = ecore_evas_get(canvas);
   if (!evas)
      return 0;

   ev = (Ecore_Event_Mouse_Button *) event;
   if (!ev)
      return 0;

   /*
    * The following doesn't work for some reason
    * so we do it manually.
    obj = evas_object_top_at_pointer_get(evas);
    */
   EINA_LIST_FOREACH(effects_windows, l, fx)
   {
      if (fx->x <= ev->x &&
	  fx->y <= ev->y && fx->x + fx->w >= ev->x && fx->y + fx->h >= ev->y)
	{
	   fx_under_mouse = fx;
	}
   }

   if (fx_under_mouse)
     {
	evas_object_raise(fx_under_mouse->img);
	e_border_raise(fx_under_mouse->border);
	e_border_focus_set_with_pointer(fx_under_mouse->border);
     }

   e_effects_revert();
   return 1;
}

static int
_e_effects_cb_mouse_wheel(void *data, int type, void *event)
{
   e_effects_revert();
   return 1;
}

static int
_e_effects_animate_cb(void *data)
{
   Eina_List *fxs = NULL;

   Eina_List *l;

   E_Effects_Object *fx;

   int done = 1;

   double delta = ecore_loop_time_get() - anim_time;

   if (!(fxs = data))
      return 1;

   if (delta > anim_length)
     {
	EINA_LIST_FOREACH(fxs, l, fx)
	{
	   e_effects_object_move_resize(fx,
					fx->dst_x, fx->dst_y,
					fx->dst_w, fx->dst_h);
	}

	if (revert)
	  {
	     _e_revert_done();
	     revert = 0;
	  }

	anim = NULL;
	return 0;
     }
   else
     {
	EINA_LIST_FOREACH(fxs, l, fx)
	{
	   int x = fx->src_x + (delta * fx->delta_x);

	   int y = fx->src_y + (delta * fx->delta_y);

	   int w = fx->src_w + (delta * fx->delta_w);

	   int h = fx->src_h + (delta * fx->delta_h);

	   e_effects_object_move_resize(fx, x, y, w, h);
	}
     }

   return 1;
}

static int
_e_revert_done()
{
   E_Effects_Object *data;

   if (ref_zone)
     {
	/* Restore the background to the zone. */
	if (zone_evas)
	  {
	     evas_object_release(ref_zone->bg_object, 0);
	     evas_object_inject(ref_zone->bg_object, zone_evas);
	     zone_evas = NULL;
	  }

	e_object_unref(E_OBJECT(ref_zone));
	ref_zone = NULL;
     }

   if (canvas)
     {
	e_canvas_del(canvas);
	ecore_evas_free(canvas);
	canvas = NULL;
     }

   /* Restore inputs */
   while (handlers)
     {
	ecore_event_handler_del(handlers->data);
	handlers = eina_list_remove_list(handlers, handlers);
     }

   e_grabinput_release(input_window, input_window);
   input_window = 0;

   /* Restore the original windows */
   while (effects_windows)
     {
	data = effects_windows->data;
	effects_windows = eina_list_remove(effects_windows, data);

	/* free memory */
	e_object_del(E_OBJECT(data));
     }
}
