#include <e.h>
#include <Ecore_Evas.h>
#include <Ecore.h>
#include <Ecore_X.h>
#include "e_effects_object.h"

static void _e_effects_object_free(E_Effects_Object * fx);
static void _e_effects_object_draw(E_Effects_Object * fx);
static int _e_effects_x_error(Display * dpy, XErrorEvent * xerror);
static void _native_surface_update(E_Effects_Object * fx);
static int _damage_cb(void *data, int type, void *event);

EAPI E_Effects_Object *
e_effects_object_new(Ecore_Evas * ee, E_Border * border, int x, int y,
		     int w, int h)
{
   E_Effects_Object *fx;

   XSetErrorHandler(_e_effects_x_error);

   fx = E_OBJECT_ALLOC(E_Effects_Object, E_EFFECTS_OBJECT_TYPE,
		       _e_effects_object_free);

   if (!fx)
      return NULL;

   fx->ecore_evas = ee;
   fx->visible = 0;
   fx->x = x;
   fx->y = y;
   fx->w = w;
   fx->h = h;

   fx->evas = ecore_evas_get(fx->ecore_evas);
   ecore_evas_name_class_set(fx->ecore_evas, "E", "e_effects_object");
   ecore_evas_title_set(fx->ecore_evas, "E Effects");

   fx->border = border;
   fx->img = evas_object_image_add(fx->evas);
   evas_object_image_filled_set(fx->img, 1);
   ecore_x_composite_redirect_window(fx->border->win,
				     ECORE_X_COMPOSITE_UPDATE_AUTOMATIC);
   fx->pixmap = ecore_x_composite_name_window_pixmap_get(fx->border->win);
   fx->damage = ecore_x_damage_new(border->win,
				   ECORE_X_DAMAGE_REPORT_RAW_RECTANGLES);
   fx->handlers = eina_list_append(fx->handlers,
				   ecore_event_handler_add
				   (ECORE_X_EVENT_DAMAGE_NOTIFY, _damage_cb,
				    fx));

   return fx;
}

EAPI void
e_effects_object_show(E_Effects_Object * fx)
{
   E_OBJECT_CHECK(fx);
   E_OBJECT_TYPE_CHECK(win, E_EFFECTS_object_TYPE);
   if (fx->visible)
      return;
   fx->visible = 1;
   ecore_evas_show(fx->ecore_evas);

   _native_surface_update(fx);
   evas_object_show(fx->img);
}

EAPI void
e_effects_object_hide(E_Effects_Object * fx)
{

   E_OBJECT_CHECK(fx);
   E_OBJECT_TYPE_CHECK(fx, E_EFFECTS_object_TYPE);
   if (!fx->visible)
      return;
   fx->visible = 0;
   evas_object_hide(fx->img);
}

EAPI void
e_effects_object_move(E_Effects_Object * fx, int x, int y)
{
   E_OBJECT_CHECK(fx);
   E_OBJECT_TYPE_CHECK(fx, E_EFFECTS_object_TYPE);

   if ((fx->x == x) && (fx->y == y))
      return;
   fx->x = x;
   fx->y = y;
   evas_object_move(fx->img, fx->x, fx->y);
}

EAPI void
e_effects_object_resize(E_Effects_Object * fx, int w, int h)
{
   E_OBJECT_CHECK(fx);
   E_OBJECT_TYPE_CHECK(fx, E_EFFECTS_object_TYPE);
   if ((fx->w == w) && (fx->h == h))
      return;
   fx->w = w;
   fx->h = h;

   evas_object_resize(fx->img, fx->w, fx->h);
}

EAPI void
e_effects_object_move_resize(E_Effects_Object * fx, int x, int y, int w, int h)
{
   E_OBJECT_CHECK(fx);
   E_OBJECT_TYPE_CHECK(fx, E_EFFECTS_object_TYPE);
   if ((fx->x == x) && (fx->y == y) && (fx->w == w) && (fx->h == h))
      return;
   fx->x = x;
   fx->y = y;
   fx->w = w;
   fx->h = h;

   evas_object_move(fx->img, fx->x, fx->y);
   evas_object_resize(fx->img, fx->w, fx->h);
}

static void
_e_effects_object_free(E_Effects_Object * fx)
{
   Ecore_Event_Handler *h;

   ecore_x_composite_unredirect_window(fx->border->win,
				       ECORE_X_COMPOSITE_UPDATE_AUTOMATIC);
   EINA_LIST_FREE(fx->handlers, h) ecore_event_handler_del(h);
   ecore_x_damage_free(fx->damage);
   free(fx);
}

static void
_e_effects_object_draw(E_Effects_Object * fx)
{
   _native_surface_update(fx);
}

static int
_e_effects_x_error(Display * dpy, XErrorEvent * xerror)
{
   char buf[PATH_MAX];

   if (XGetErrorText(dpy, xerror->error_code, buf, PATH_MAX))
     {
	fprintf(stderr, "X ERROR %5d: %s\n", xerror->error_code, buf);
     }
   return 1;
}

static void
_native_surface_update(E_Effects_Object * fx)
{
   int iw, ih;
   int pw, ph;
   XImage *xim;

   evas_object_image_size_get(fx->img, &iw, &ih);
   ecore_x_drawable_geometry_get(fx->pixmap, NULL, NULL, &pw, &ph);
   xim = XGetImage(ecore_x_display_get(), fx->pixmap, 0, 0,
		   pw, ph, AllPlanes, ZPixmap);
   evas_object_size_hint_request_set(evas_object_smart_parent_get(fx->img), pw,
				     ph);

   evas_object_image_data_update_add(fx->img, 0, 0, iw, ih);
   if (!xim)
     {
	printf("No X Image :-(\n");
	return;
     }

   if ((pw != iw) || (ph != ih))
     {
	evas_object_image_alpha_set(fx->img, 0);
	evas_object_image_size_set(fx->img, pw, ph);
	evas_object_image_size_get(fx->img, &iw, &ih);
     }

   if ((iw < 1) || (ih < 1))
     {
	printf("Size < 1\n");
	XDestroyImage(xim);
	return;
     }
   else
     {
	int *data = NULL;

	int size = sizeof(int) * iw * ih;

	data = evas_object_image_data_get(fx->img, 1);

	if (data)
	  {
	     /* TODO
	      * Make sure they are both ARGB instead of assuming if
	      * they are not, do something smart like a convertion.
	      */
	     memcpy(data, xim->data, size);
#if 0
	     Evas_Native_Surface ns;

	     ns.mode = SOFTWARE;
	     ns.data.rgba = xim->data;
	     ns.w = fx->w;
	     ns.h = fx->h;

	     native_surface_set(fx->img, ns);

	     evas_object_resize(fx->img, fx->w, fx->h);
	     evas_object_image_fill_set(fx->img, 0, 0, fx->w, fx->h);
#endif
	  }
	XDestroyImage(xim);
     }
}

static int
_damage_cb(void *data, int type, void *event)
{
   E_Effects_Object *fx;

   Ecore_X_Event_Damage *ev;

   if (!(fx = data))
      return 1;
   if (!(ev = event))
      return 1;
   if (fx->border->win != ev->drawable)
      return 1;
   _native_surface_update(fx);
   return 1;
}
